import os
from os import path
from lxml import etree
import regex

HERE = path.abspath(path.split(__file__)[0])

re_urn = regex.compile(r'^[^\.]+\.[^\.]+\.[^\-]+-([^\.]+)(?=\.xml$)')

with open(path.join(HERE, 'ca_template.xml')) as f:
    template = f.read()

def parse_urn(fname):
    match = re_urn.search(path.split(fname)[1])
    if match:
        return match.group(), match.group(1)
    else:
        raise Exception(f'Urn {urn} not of correct format')

def to_string(element):
    content = etree.tostring(element, encoding='unicode')
    content = regex.sub(fr'xmlns(?:\:[a-zA-Z]+?)?=".+?"\s*', '', content) # remove all NS
    return content

def recursive_update(dct, passage, value):
    n = passage[0]
    if len(passage) == 1:
        if n in dct.keys():
            dct[n].append(value)
        else:
            dct[n] = [value]
    else:
        if n in dct.keys():
            dct[n] = recursive_update(dct[n], passage[1:], value)
        else:
            dct[n] = recursive_update(dict(), passage[1:], value)
    return dct

def collect_apps(apps, re_passage, NS):
    print(f"{len(apps)} apps found")
    ca_dict = dict()
    not_matched = 0
    for app in apps:
        loc = app.attrib['loc']
        # print(loc)
        match = re_passage.search(loc)
        if match:
            passage = match.groups()
            app_string = to_string(app)
            ca_dict = recursive_update(ca_dict,passage,app_string)
        else:
            not_matched += 1
    print(f"{len(apps) - not_matched} of which loc attribute succesfully interpreted")
    return ca_dict

def format_critapp(ca_dict, levels):
    ca_string = ''
    tag = levels[0]
    for key,value in ca_dict.items():
        tp_type = ' type="textpart"' if tag != 'l' else ''
        if type(value) == list: # len(levels) == 1
            apps_string = '\n'.join(value)
            ca_string += f'<{tag}{tp_type} n="{key}">\n<listApp>\n{apps_string}\n</listApp>\n</{tag}>\n'
        elif type(value) == dict:
            ca_string += f'<{tag}{tp_type} n="{key}">\n{format_critapp(value,levels[1:])}\n</{tag}>\n'
    return ca_string

def create(fname, ca_ext, data_dir='.', lang='', app_type=None):
    urn, extension = parse_urn(fname)
    tree = etree.parse(path.join(data_dir, fname))
    root = tree.getroot()
    NS = f'{{{root.nsmap[None]}}}'
    if not lang:
        lang = etree.ETXPath(f'//{NS}text/{NS}body/{NS}div/@xml:lang')(root)[0]
    print(f"Lanugage: {lang}")

    # find refsDecl
    refsDecl = to_string(etree.ETXPath(f'//{NS}refsDecl')(root)[0])

    # Catch first cRefPattern
    cRef = etree.ETXPath(f'//{NS}refsDecl/{NS}cRefPattern/@replacementPattern')(root)[0]
    levels = regex.findall(r'tei:([a-zA-Z]+)\[@n=[\'"]\$[0-9]+[\'"]\]',cRef)
    print(f"cRefPattern: {', '.join(levels)}")

    # find apps
    passage = r'\.'.join(r'(\w+)' for _ in range(len(levels)))
    urn_wo_ext = regex.sub('-.+?$','',urn)
    print(fr'Searching for {urn_wo_ext}-[\w\-]+?:{passage}')
    re_passage = regex.compile(fr'{urn_wo_ext}-[\w\-]+?:{passage}')
    app_xpath = f'//{NS}app[@loc]'
    if app_type:
        app_xpath = f'//{NS}listApp[@type="{app_type}"]' + app_xpath
    apps = etree.ETXPath(app_xpath)(root)
    ca_dict = collect_apps(apps,re_passage,NS)

    ca_string = format_critapp(ca_dict, levels)

    # Create new file
    new_urn = regex.sub(f'{extension}$', f'{ca_ext}', urn)
    new_fname = regex.sub(f'{extension}(?=.xml$)', f'{ca_ext}', fname)
    content = template.format(content=ca_string, id=new_urn, refsDecl=refsDecl, lang=lang)

    # Indentation
    tree = etree.fromstring(content)
    etree.indent(tree)

    with open(path.join(data_dir, new_fname), 'w') as f:
        f.write(etree.tostring(tree, encoding="unicode"))



import logging
import os
from contextlib import contextmanager
from copy import copy
from dataclasses import dataclass
from itertools import chain
from pathlib import Path

import argparse
from lxml import etree
from anytree import Node, LevelOrderGroupIter, RenderTree
from lxml.etree import Element, dump, ParseError
from typing import List

TEI_XPATH = "/tei:TEI/"
CTS_VERSION_XPATH = "tei:text/tei:body/tei:div[@type][@n]"
CTS_TEXTPART_XPATHS = ["tei:div[@type][@n]", "tei:l[@n]", "tei:ab/tei:l[@n]"]
NSMAP = {"tei": "http://www.tei-c.org/ns/1.0"}


class RefNode(Node):
    element: Element
    n: str
    type: str = None
    subtype: str = None
    xpath_match: str = None


class TextpartNode(RefNode):
    pass


@dataclass
class TextpartPath:
    segments: List[str]
    type: str


def is_textpart(node):
    return isinstance(node, TextpartNode)


def findall(el, xpath):
    """Find all child elements that match xpath query"""
    return el.findall(xpath, namespaces=NSMAP)


def build_node(element, *, parent=None, node_type=RefNode, xpath_match=None):
    type_ = element.attrib.get("type") or node_type.type
    subtype_ = element.attrib.get("subtype") or node_type.subtype

    return node_type(
        name=f"{type_} ({subtype_})" if subtype_ else type_,
        element=element,
        parent=parent,
        type=type_ or "textpart",
        subtype=subtype_ or "line",
        n=element.attrib.get("n"),
        xpath_match=xpath_match,
    )


def build_ref_tree(el):
    work = build_node(el, parent=None)

    for version_el in findall(el, f"./{CTS_VERSION_XPATH}"):
        version = build_node(version_el, parent=work)
        build_textpart_tree(version_el, parent=version)

    return work


def build_textpart_tree(el, parent):
    for xpath_match, textpart_el in find_textparts(el):
        textpart = build_node(
            element=textpart_el,
            parent=parent,
            node_type=TextpartNode,
            xpath_match=xpath_match,
        )
        build_textpart_tree(textpart_el, parent=textpart)


def find_textparts(el):
    return chain(
        *(
            [(path, element) for element in findall(el, f"./{path}")]
            for path in CTS_TEXTPART_XPATHS
        )
    )


def textpart_levels(tree):
    levels, current_level = [], []

    for group in filter(None, LevelOrderGroupIter(tree, filter_=is_textpart)):
        current_level.append(group[0])
        levels.append(copy(current_level))

    return levels


def textpart_paths(tree):
    for ref in textpart_levels(tree):
        segments, type_ = [], None

        for i, node_ in enumerate(ref, start=1):
            type_ = node_.subtype
            segments.append(node_.xpath_match.replace("[@n]", f"[@n='${i}']"))

        yield TextpartPath(segments=segments, type=type_)


def debug_tree(tree):
    for pre, fill, node in RenderTree(tree):
        print("%s%s" % (pre, node.name))


def build_refs_decl(tree, path_root):
    element = etree.Element("refsDecl", {"n": "CTS"})

    for path in reversed([path for path in textpart_paths(tree)]):
        attrib = {
            "n": path.type,
            "matchPattern": r"\.".join([r"(\w+)"] * len(path.segments)),
            "replacementPattern": f"#xpath({os.path.join(path_root, *path.segments)})",
        }
        element.append(etree.Element("cRefPattern", attrib))

    return element


def update_refsdecl(tree, refs_decl, path):
    encoding_desc = tree.find("//tei:encodingDesc", NSMAP)
    if encoding_desc is None:
        raise Exception("Missing enveloping element 'encodingDesc'")

    existing_refsdecl = encoding_desc.find("./tei:refsDecl", NSMAP)
    if existing_refsdecl is not None:
        encoding_desc.remove(existing_refsdecl)

    encoding_desc.append(refs_decl)

    with open(path, "wb") as writer:
        tree.write(writer, encoding="utf-8")


@contextmanager
def read_xml(path):
    with open(path, "rb") as reader:
        try:
            root_node = etree.fromstring(reader.read())
            yield etree.ElementTree(root_node)
        except ParseError as e:
            logging.exception(Exception(f"Could not parse: {path}", e))
            yield etree.ElementTree()


def xml_paths(path):
    for subdir, dirs, files in os.walk(path):
        yield from (
            os.path.join(subdir, file) for file in files if file.endswith(".xml")
        )


def is_tei_xml(tree):
    return tree.getroot() is not None and tree.find("//tei:*", NSMAP) is not None


def generate_for_file(path, update):
    with read_xml(path) as tree:
        # filter out all non-tei files
        if not is_tei_xml(tree):
            return

        element = build_refs_decl(
            tree=build_ref_tree(el=tree.getroot()),
            path_root=os.path.join(TEI_XPATH, CTS_VERSION_XPATH),
        )

        if update:
            update_refsdecl(tree, element, path)
            print(f"Succesfully updated {path}")

        return element


def generate_for_path(path, update):
    path_ = Path(path)
    for path_ in ([path_] if path_.is_file() else xml_paths(path_)):
        element = generate_for_file(path_, update)
        if element is not None:
            yield element


def generate(args):
    for refsdecl in generate_for_path(path=args.path, update=args.update):
        if not args.update:
            dump(refsdecl)


def parse_args():
    parser = argparse.ArgumentParser(
        description="A tool to dynamically generate refsDecl definitions for TEI files"
    )
    parser.add_argument("path", help="Path to TEI file")
    parser.add_argument(
        "--update",
        action="store_true",
        help="Updates the file with the newly generated refsDecl",
    )
    return parser.parse_args()


if __name__ == "__main__":
    generate(args=parse_args())



import regex
from lxml import etree
from os import path
from xml.sax.saxutils import unescape, escape

__all__ = ['LanguageNotSupported', 'tag_script', 'tag_script_from_file']

class LanguageNotSupported(Exception):
    pass

RANGE = {
    'Arab': '\u0606-\u060b\u060e\u060f\u061c-\u064a\u0660-\u066a'\
            '\u066d-\u066f\u0671-\u06d5\u06de\u06e5\u06e6\u06e9\u06ee-\u06ff',
    'Copt': '\u03e2-\u03ef\u2c80-\u2cff\u2e0c\u2e0d',
    'Hebr': '\u05d0-\u05ff\ufb20-\ufb4f',
    'Latn': 'a-zA-Z0-9\u00c0-\u00ff\u1e00-\u1ef9',
    'Cyrl': '\u0400-\u0482\u048a-\u04ff\ua640-\ua66e\ua67e-\ua69d'
}

COMBINING_PUNCT = {
    'Arab': '\u0600-\u0605\u060c\u060d\u0610-\u061b\u064b-\u065f\u066b\u066c'\
            '\u0670\u06d6-\u06dd\u06df-\u06e4\u06e7\u06e8\u06ea-\u06ed',
    'Copt': '\u0300-\u0361',
    'Hebr': '\u0590-\u05cf\ufb1d-\ufb1f',
    'Latn': '{}.,\(\)!:?;',
    'Cyrl': '\u0301\u20dd\u0483-\u0489\u2de0-\u2dff\ua66f-\ua67d\ua69e\ua69f'
}

DIRECTION_MARKERS = '\u200e\u202a-\u202e'

RE_STR = {
    script: fr'[{DIRECTION_MARKERS}]*[{RANGE[script]}][\s{RANGE[script]}{COMBINING_PUNCT[script]}{DIRECTION_MARKERS}]*(?<!\s)'
    for script in RANGE.keys()
}

# Language specific additions
RE_STR['Latn'] = '(?<!&#?[a-zA-Z0-9]*)' + RE_STR['Latn'] # Avoid escaped xml chars

RE = {script: regex.compile(RE_STR[script]) for script in RANGE.keys()}

DEFAULT_LCS = {
    'Arab': 'ar-Arab',
    'Copt': 'cop-Copt',
    'Hebr': 'he-Hebr',
    'Latn': 'la-Latn',
    'Cyrl': 'cu-Cyrl'
}

AVAILABLE_SCRIPTS = list(DEFAULT_LCS.keys())

def tag_script(string, script, language_code = '', escape_xml = True):
    if escape_xml:
        string = escape(string)
    if script not in AVAILABLE_SCRIPTS:
        raise LanguageNotSupported(
            f'Language "{script}" not (yet) supported, please use one of: '
            + ', '.join(AVAILABLE_SCRIPTS)
        )
    if not language_code:
        language_code = DEFAULT_LCS[script]
    return RE[script].sub(
        f'<foreign xml:lang="{language_code}">\g<0></foreign>', string
    )

XML_NS = '{http://www.w3.org/XML/1998/namespace}'

def tag_script_from_file(fname, script, language_code = ''):
    if not language_code:
        language_code = DEFAULT_LCS[script]
    tree = etree.parse(fname)
    root = tree.getroot()
    NS = f'{{{root.nsmap[None]}}}' if None in root.nsmap.keys() else ''
    for string in etree.ETXPath(f'//{NS}body//text()')(root):
        parent = string.getparent()
        new_content = escape(str(string))
        # Check first if first ancestor with xml:lang doesn't already declares
        # the language code to be inserted
        if string.is_text:
            lang_parent = parent.xpath(f'ancestor-or-self::*[@xml:lang][1]')
        elif string.is_tail:
            lang_parent = parent.xpath(f'ancestor::*[@xml:lang][1]')

        if (not lang_parent) or (lang_parent[0].attrib[f'{XML_NS}lang'] != language_code):
            new_content = tag_script(new_content, script, language_code, escape_xml = False)

        # Create XML tree from tagged content and append to existing tree
        try:
            new_xml_root = etree.fromstring(f'<root>{new_content}</root>')
        except etree.XMLSyntaxError:
            raise Exception(new_content)

        if string.is_text:
            parent.text = new_xml_root.text
            for i,el in enumerate(new_xml_root.getchildren()):
                parent.insert(i, el)
        elif string.is_tail:
            parent.tail = new_xml_root.text
            grandparent = parent.getparent()
            i_parent = grandparent.index(parent)
            for i,el in enumerate(new_xml_root.getchildren()):
                grandparent.insert(i_parent + 1 + i, el)

    with open(fname, 'w', encoding = 'utf-8') as f:
        f.write(etree.tostring(tree, encoding = 'unicode'))



from markdown.postprocessors import Postprocessor
from markdown.extensions import Extension
from markdown import markdown

from lxml import etree
import re
from namedentities import unicode_entities

__all__ = ['TEIPostprocessor', 'ToTEI', 'md2tei']


class TEIPostprocessor(Postprocessor):

    def __init__(self, md, indent=False, with_root=False):
        self.indent = indent
        self.with_root = with_root
        Postprocessor.__init__(self, md)

    def _replace_tag(self, old_tag, new_tag, clear_attrib=False, **attribs):
        for el in self.tree.xpath(f'//{old_tag}'):
            el.tag = new_tag

            new_attrib = {}
            for key, value in attribs.items():
                if value.startswith('@'):
                    new_attrib[key] = el.attrib.pop(value[1:], '')
                else:
                    new_attrib[key] = value

            if clear_attrib:
                el.attrib = {}
            el.attrib.update(new_attrib)

    def run(self, text, indent_unit='  '):
        root_tag = 'root'

        # A dirty namespace hack
        text = unicode_entities(text)
        text = text.replace('{http://www.w3.org/XML/1998/namespace}', 'xml:')

        self.tree = etree.fromstring(f'<{root_tag}>{text}</{root_tag}>')

        # heads (from old script, not sure if this is actually encountered, AvD)
        for el in self.tree.xpath('//head'):
            el.tag = 'p'
            head = etree.SubElement(el, 'seg')
            head.attrib['type'] = 'head'

            # Move content of over to <seg>
            head.text = el.text
            el.text = ''
            for child in el.getchildren():
                head.append(child)

        # tables
        for table in self.tree.xpath('//table'):
            new_table = etree.Element('table')

            if table.xpath('thead'):
                row = etree.SubElement(new_table, 'row')
                row.attrib['role'] = 'label'
                for cell in table.xpath('thead/tr/th'):
                    cell.tag = 'cell'
                    row.append(cell)

            for table_row in table.xpath('*[not(self::thead)]/tr'):
                row = etree.SubElement(new_table, 'row')
                for cell in table_row.xpath('td'):
                    cell.tag = 'cell'
                    row.append(cell)

            new_table.tail = table.tail

            container = table.getparent()
            container.insert(container.index(table) + 1, new_table)
            del container[container.index(table)]

        self._replace_tag('em', 'hi', rend='italic')
        self._replace_tag('i', 'hi', rend='italic')
        self._replace_tag('strong', 'hi', rend='bold')
        self._replace_tag('b', 'hi', rend='bold')
        self._replace_tag('sup', 'hi', rend='superscript')
        self._replace_tag('sub', 'hi', rend='subscript')
        self._replace_tag('small', 'hi', rend='smallcaps')
        for i in range(1, 7):
            self._replace_tag(f'h{i}', 'head')
        self._replace_tag('br', 'lb')
        self._replace_tag('ol', 'list', rend='numbered')
        self._replace_tag('ul', 'list', rend='bulleted')
        self._replace_tag('li', 'item')
        self._replace_tag('blockquote', 'quote')
        self._replace_tag('a', 'ref', target='@href')
        self._replace_tag('img', 'graphic', n='@alt', url='@src')

        if self.indent:
            etree.indent(self.tree, space=indent_unit)

        new_text = etree.tostring(self.tree, encoding='unicode')

        if not self.with_root:
            # Remove wrapped root element
            new_text = re.sub(rf'^\s*<{root_tag}>|</{root_tag}>\s*$', '', new_text)
            if self.indent:
                new_text = new_text.replace('\n' + indent_unit, '\n')

        return new_text


class ToTEI(Extension):

    def __init__(self, **kwargs):
        self.config = {
            'indent': [False, ''],
            'with_root': [False, '']
        }
        Extension.__init__(self, **kwargs)

    def extendMarkdown(self, md):
        configs = self.getConfigs()
        md.postprocessors.register(TEIPostprocessor(md, configs['indent'], configs['with_root']), 'to_tei', 0)


def md2tei(text):
    return markdown(text, extensions=[ToTEI()])



import unicodedata
import json
import string
import os
try:
    from cltk.corpus.greek.beta_to_unicode import Replacer
except ModuleNotFoundError:
    CLTK_NOT_FOUND = 1
else:
    CLTK_NOT_FOUND = 0

FILE_DIR = os.path.split(__file__)[0]
UNI_BETA_DICT = {}
for fname in ['UNI_BETA_LOWER', 'UNI_BETA_UPPER']:
    with open(os.path.join(FILE_DIR, f'{fname}.json')) as f:
        UNI_BETA_DICT.update(json.load(f))

UNI_BETA_TRANS = str.maketrans(UNI_BETA_DICT)

# str.lower(), but only for latin alphabet
LATIN_UPPER_TRANS = str.maketrans(string.ascii_lowercase, string.ascii_uppercase)
LATIN_LOWER_TRANS = str.maketrans(string.ascii_uppercase, string.ascii_lowercase)

def decompose(string, save_as=''):
    decomp = []
    for char in string:
        if len(hex(ord(char))[2:]) <= 4:
            hex_str = 'U+{:>4}'.format(hex(ord(char))[2:]).replace(' ','0')
        else:
            hex_str = hex(ord(char)).replace('0x','U+')

        try:
            name = unicodedata.name(char)
        except ValueError:
            name = ''

        decomp.append(' '.join([char, hex_str, name]))
    if save_as:
        with open(save_as, 'w') as f:
            f.write('\n'.join(decomp))
    else:
        print('\n'.join(decomp))

def beta2uni(text_beta):
    """ Wrapper of the cltk.corpus.greek.beta_to_unicode.Replacer function """
    if CLTK_NOT_FOUND:
        print(
            'CLTK is not found in this environment. In order to use the beta2uni converter,',
            'install this package with `pip install cltk` or `pip install dh-utils[betacode]`'
        )
        return None
    text_beta = text_beta.translate(LATIN_UPPER_TRANS)
    text_uni = Replacer().beta_code(text_beta.upper())
    return text_uni

def uni2beta(text_uni, normalize=True):
    """ Inverse of beta2uni """
    if normalize:
        text_uni = unicodedata.normalize('NFC', text_uni)
    text_beta = text_uni.translate(UNI_BETA_TRANS)
    text_beta = text_beta.translate(LATIN_LOWER_TRANS)
    return text_beta



from setuptools import setup, find_packages
from codecs import open
from os import path

here = path.abspath(path.dirname(__file__))

with open(path.join(here, 'README.md')) as f:
    long_description = f.read()

setup(
    name='dh-utils',
    version='0.1.21',
    description='Python package containing various utilities relevant in the field of digital humanities.',
    packages=find_packages(),
    install_requires=['regex', 'lxml', 'anytree', 'markdown', 'namedentities'],
    extras_require={
        'betacode': ['cltk']
    },
    keywords='digital humanities utilities unicode tei',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url="https://github.com/andredelft/dh-utils",
    include_package_data=True,
    python_requires='>=3.6',
    author='André van Delft',
    author_email='andrevandelft@outlook.com',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Programming Language :: Python :: 3'
    ]
)
